package org.edg.wl.dgas.dgassim;

import java.util.*;
import java.io.*;

import org.edg.wl.dgas.dgassim.workload.pa.*;
import org.edg.wl.dgas.dgassim.workload.rb.*;
import org.edg.wl.dgas.dgassim.monitor.*;
import org.edg.wl.dgas.dgassim.fabric.*;

/**
   <code><b>AutoJobSubmSimCore</b></code> is a <code><b>Thread</b></code>
   that implememts the {@link org.edg.wl.dgas.dgassim.SimCore SimCore}
   interface. It periodically submits random jobs to a single central
   {@link org.edg.wl.dgas.dgassim.workload.rb.ResourceBroker ResourceBroker}.
   Note: The class doesn't implement the method
   <code><b>Thread.stop()</b></code> but instead keeps a flag
   (<code>runSim</code>) that is checked in {@link #run run()}.
 */
public class AutoJobSubmSimCore extends Thread implements SimCore {

    private boolean verbose;
    private boolean debug;

    private static int MSEC_BETWEEN_STDOUT, MSEC_BETWEEN_JOBS;

    private static int JOBS_PER_SUBMISSION;

    private static String PRICE_OUTFILE_PREFIX, QWT_OUTFILE_PREFIX;

    private BufferedWriter jobsubmOutFile;

    /**
       defines the maximum duration of the simulation in miliseconds.
     */
    public static long maxSimRunTime; // max. duration (in miliseconds)

    private long timeOfSimStart; // timestamp of start of simulation
    private long timeOfLastComp; // timestamp of last computation
    private long timeOfLastJob; // timestamp of last job sumbission

    private long totalCompEnergy; // computational energy requested by all
                                 // jobs during run time.
    private long totalCompEnSubmitted; // total comp. en. of all submitted jobs

    private GIS gis;
    private ResourceBroker rb;

    private int lastJobId, noJobsRefused;

    private boolean runSim;

    /**
       creates an instance.

       @param gis the {@link org.edg.wl.dgas.dgassim.monitor.GIS GIS}
       (Grid Information System) instance to which all
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
       and
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}s
       are registered.
       @param rb the
       {@link org.edg.wl.dgas.dgassim.workload.rb.ResourceBroker ResourceBroker}
       used for the resource selection process.
       @param msMaxRunTime the maximum simulation run time in miliseconds.
       @param priceOutfilePrefix the prefix for the names of the output files
       containing the price information over time (one file per
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}).
       @param qwtOutFilePrefix the prefix for the names of the output files
       containing the information about the queue wait times (queue traversal
       times) over time (one file per
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}).
       @param jobsubout the name of the logfile for job submission information.
       @param msStdout the number of miliseconds between the display of
       information to the standard output (if verbose).
       @param msJobs the number of miliseconds between job submissions.
       @param jobsPerSubm the number of jobs per submission.
       @param verbose sets the verbose flag to print information to the
       standard output.
       @param debug sets the debug flag to print additional debugging
       information to the standard output.
     */
    public AutoJobSubmSimCore(GIS gis, ResourceBroker rb, long msMaxRunTime,
			      String priceOutfilePrefix,
			      String qwtOutFilePrefix,
			      String jobsubout,
			      int msStdout, int msJobs, int jobsPerSubm,
			      boolean verbose, boolean debug) {
	super();

	timeOfLastComp = (new Date()).getTime();
	timeOfLastJob = (new Date()).getTime();

	this.gis = gis;
	this.rb = rb;

	this.verbose = verbose;
	this.debug = debug;

	PRICE_OUTFILE_PREFIX = priceOutfilePrefix;
	QWT_OUTFILE_PREFIX = qwtOutFilePrefix;

	try {
	    jobsubmOutFile = new BufferedWriter(new FileWriter(jobsubout));
	    jobsubmOutFile.write("# Job submission logfile\n"
				 + "# the boolean number at the "
				 + "beginning of each records "
				 + "indicates,\n"
				 + "# whether the job could be "
				 + "submitted or not\n");
	} catch (Exception e) {
	    System.err.println("Cannot create output file for job submission "
			       + "log");
	    System.exit(-1);
	}

	MSEC_BETWEEN_STDOUT = msStdout;
	MSEC_BETWEEN_JOBS = msJobs;
	JOBS_PER_SUBMISSION = jobsPerSubm;

	maxSimRunTime = msMaxRunTime;

	if (debug)
	    System.out.println("constructor of class AutoJobSubmSimCore");

	if (debug)
	    System.out.println("MSEC_BETWEEN_STDOUT="+MSEC_BETWEEN_STDOUT
			       +", MSEC_BETWEEN_JOBS="+MSEC_BETWEEN_JOBS
			       +", JOBS_PER_SUBMISSION="+JOBS_PER_SUBMISSION
			       +", maxSimRunTime="+maxSimRunTime);
    } // constructor

    /**
       writes the headers of the
       output files for price and QWT information.
     */
    public void initSimulation() {

	totalCompEnergy = 0;
	totalCompEnSubmitted = 0;
	lastJobId = 0;
	noJobsRefused = 0;

	ArrayList ceList = gis.getCEList();

	for (int i = 0; i < ceList.size(); i++) {

	    ComputingElement ce = (ComputingElement)ceList.get(i);
	    int id = ce.getId();

	    try {
		ce.setQWTOutFile(new BufferedWriter(new FileWriter(
						          QWT_OUTFILE_PREFIX
						          +"_CE"+id+".dat")));
		ce.getQWTOutFile().write("# current QWT(time units) of CE"+id
					 + " (as function of time(s))\n"
					 + "# time QWT\n"
					 + "0 "+ce.getCurrQWTNoComp()+"\n");
		ce.setPriceOutFile(new BufferedWriter(new FileWriter(
							  PRICE_OUTFILE_PREFIX
							  +"_CE"+id+".dat")));
		ce.getPriceOutFile().write("# current Price(GridCredits) of CE"
					   + id + " (as function of time(s))\n"
					   + "# time Price\n"
					   + "0 "+ce.getPrice()+"\n");
	    } catch (Exception e) {
		System.err.println("Cannot create output files for CE"+id);
		System.exit(-1);
	    }
	}
    }

    /**
       starts the <code><b>Thread</b></code>
       by calling {@link #start start()}.
     */
    public void runSimulation() {

	timeOfSimStart = (new Date()).getTime();

	// start AutoJobSubmSimCore
	start();
    }

    /**
       inherited from <code><b>Thread</b></code>. Starts this thread and the
       threads of all
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}
       instances registered to the Grid Information
       System ({@link org.edg.wl.dgas.dgassim.monitor.GIS GIS}). 
    */
    public void start() {
	runSim = true;
	super.start();

	System.out.println("Starting simulation with AutoJobSubmSimCore for "
			   + maxSimRunTime + " ms = "
			   + ((double)maxSimRunTime/60000) + " min.");
	if (verbose) System.out.println("Starting AutoJobSubmSimCore thread");

	// starting all PA threads
	ArrayList paList = gis.getPAList();
	for (int i = 0; i < paList.size(); i++)
	    ((PriceAuthority)paList.get(i)).start();
    }


    /**
       stops the simulation by
       stopping all threads. Note: The class doesn't implement the method
       <code><b>Thread.stop()</b></code> but instead keeps a flag
       (<code>runSim</code>)that is periodically checked in {@link #run run()}.
       This method sets <code>runSim</code> to <code>false</code>.
     */
    public void stopSimulation() {
	runSim = false;

	// stopping PA threads
	ArrayList paList = gis.getPAList();
	for (int i = 0; i < paList.size(); i++)
	    ((PriceAuthority)paList.get(i)).stopPA();

	long currTime = (new Date()).getTime();
	long seconds = (currTime - timeOfSimStart) / 1000;
	System.out.println("stopping at " + seconds + " sec.\n"
			   + lastJobId
			   + " job submission requests with a total comp. "
			   + "energy of " + totalCompEnergy + " UCE = "
			   + (totalCompEnergy/seconds) + " UCE/s.\n"
			   + (lastJobId-noJobsRefused)
			   + " requests satisfied for a total comp. energy of "
			   + totalCompEnSubmitted + " UCE = "
			   + (totalCompEnSubmitted/seconds) + " UCE/s.");
    }

    /**
       inherited from <code><b>Thread</b></code>. Periodically submits jobs to
       the
       {@link org.edg.wl.dgas.dgassim.workload.rb.ResourceBroker ResourceBroker}.
       The number of jobs per submission, the time interval between job
       submissions, and the time interval between displaying information on
       the price and the QWT of all registered
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
       to the standard output are set when calling the constructor (see
       {@link #AutoJobSubmSimCore AutoJobSubmSimCore()}. Note: The class
       doesn't implement the method <code><b>Thread.stop()</b></code> but
       instead keeps a flag (<code>runSim</code>) that is periodically checked
       in <code><b>run()</b></code>. A while-loop is executed  as long as
       <code>runSim</code> is <code>true</code>.
     */
    public void run() {

	long currTime;

	int jobID = 1;

	int price, cQWT;

	ArrayList ces = gis.getCEList();
	ArrayList ceIDs = gis.getCEIDList();

	while(runSim) {


	    // first check if we need to submit a job:
	    currTime = (new Date()).getTime();

	    // maximum run time reached?
	    if (maxSimRunTime != 0) // limit specified
		if (currTime >= maxSimRunTime + timeOfSimStart)
		    stopSimulation();

	    // submit jobs?
		if ( (currTime - timeOfLastJob) >= MSEC_BETWEEN_JOBS) {

		    boolean submitted;

		    for (int j = 0; j < JOBS_PER_SUBMISSION; j++) {

			// creating at random one of the job types
			Job newJob = gis.createNewJob();

			// trace consumed comp. energy
			totalCompEnergy += newJob.compEnergy();

			submitted = rb.submitJob(newJob);


			// write to job submission logfile:
			try {
			    if (submitted)
				jobsubmOutFile.write("1 ");
			    else
				jobsubmOutFile.write("0 ");

			    // in any case:
			    jobsubmOutFile.write("timestamp=" + currTime +
						 " jobId=" + newJob.edgJobId()
						 + " type=" + newJob.jobType()
						 + " compEn="
						 + newJob.compEnergy()
						 + " reqCPU="
						 + newJob.minCPUPower());

			    if (submitted) 
				jobsubmOutFile.write(" CEId="
						     + newJob.executingCE());

			    if (submitted)
				totalCompEnSubmitted += newJob.compEnergy();
			    else
				noJobsRefused++;

			    // terminate record
			    jobsubmOutFile.newLine();
			    jobsubmOutFile.flush();

			}
			catch (Exception e) {
			    System.err.println("Cannot write to job submission"
					       + " logfile.");
			}

			lastJobId = newJob.edgJobId();

			if (verbose && !submitted)
			    System.out.println("Job " + newJob.edgJobId()
					       + " of type " + newJob.jobType()
					       + " NOT "
					       + "submitted: CPU power not "
					       + "sufficient or queues full.");
		    }
		    timeOfLastJob = (new Date()).getTime();
		}

	    // now monitor QWTs and prices ...

	    currTime = (new Date()).getTime();
	    if ( (currTime - timeOfLastComp) >= MSEC_BETWEEN_STDOUT ) {
		for (int i = 0; i < ceIDs.size(); i++) {
		    ComputingElement ce = (ComputingElement)
			      gis.getCE(((Integer)ceIDs.get(i)).intValue());
		    price = ce.getPA().getPrice(ce.getId());
		    cQWT = ce.getCurrQWT();
		    if (verbose) {
			System.out.print("CE " + i + ": p=" + price
					 + ", cQWT=" + cQWT + "; ");
			if (i == ceIDs.size()-1)
			System.out.print("\n");
		    }
		}
		timeOfLastComp = currTime;
	    }

	    if (!runSim) {
		try {
		    jobsubmOutFile.close();
		}
		catch (Exception e) {
		    System.err.println("Cannot close job submission"
				       + " logfile.");
		}
	    }

	} // while(runSim)...

    }

}
