package org.edg.wl.dgas.dgassim;

import java.util.ArrayList;
import org.edg.wl.dgas.dgassim.fabric.*;

/**
   represents a job and its job description. A limited number of job
   requirements are simulated (minimum processing power and queue types).
   Derived classes might be used to implement further requirements (e.g. OS
   type). If deriving the <code><b>Job</b></code> class for this purpose you
   may need to overwrite its method {@link #matchCE matchCE(ce)} as well.
 */
public class Job {

    private int edgJobId;
    /**
       the default job ID for job types (prototypes) is
       <code><b>Integer.MIN_VALUE</b> (-2147483648)</code>.
     */
    public static final int PROTOTYPE = Integer.MIN_VALUE;

    private int jobType;

    private int compEnergy;
    private int minCPUPower;

    private int execCEId; // id of executing CE.
    /**
       the default ID of the executing
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       is <code><b>Integer.MIN_VALUE</b> (-2147483648)</code> to indicate that
       the job has not yet been submitted.
     */
    public static final int NO_EXEC_CE = Integer.MIN_VALUE;

    private int minJobCompEn, maxJobCompEn;
    private int minJobCPUReq, maxJobCPUReq;

    private ArrayList requiredQTypes;

    private int price;

    /**
       creates a new job type randomly choosing the omputational energy
       "consumed" and the minimum processing power required between the
       specified limits. Although this constructor may be used to create a
       single job (without a job ID, see {@link #PROTOTYPE Job.PROTOTYPE}),
       it is recommended to use it only for generating job TYPES. To generate
       a single random job of a specific type call the method
       {@link #newInstance newInstance(id)} of that job type.
       @param type a numerical ID for the job type, should be unique.
       @param minCompEn the minimum value for the computational energy
       "consumed" by a generated random job of this type. Measured in
       <i>units of computational energy (UCE)</i>.
       @param maxCompEn the maximum value for the computational energy
       "consumed" by a generated random job of this type. Measured in
       <i>units of computational energy (UCE)</i>.
       @param minCPUReq the minimum value for the processing power required by
       a generated random job of this type. Measured in <i>UCE per time unit</i>.
       @param maxCPUReq the maximum value for the processing power required by
       a generated random job of this type. Measured in <i>UCE per time unit</i>.
       @param reqQT an <code><b>ArrayList</b></code> of ID strings of all queue
       types to which a job of that type may be submitted. Queue ID strings are
       defined in the configuration file.
     */
    public Job(int type, int minCompEn, int maxCompEn, int minCPUReq,
	       int maxCPUReq, ArrayList reqQT) {

	edgJobId = PROTOTYPE;
	jobType = type;
	execCEId = NO_EXEC_CE;

	compEnergy =
	    (int) ((double)(maxCompEn - minCompEn)*(double)Math.random() + 0.5)
	          + minCompEn;

	minJobCompEn = minCompEn;
	maxJobCompEn = maxCompEn;
	minJobCPUReq = minCPUReq;
	maxJobCPUReq = maxCPUReq;

	minCPUPower =
	    (int) ((double)(maxCPUReq - minCPUReq)*(double)Math.random() + 0.5)
	          + minCPUReq;

	requiredQTypes = reqQT;
    }

    /**
       returns the job's ID.
       @return {@link #PROTOTYPE Job.PROTOTYPE} if this is a job type
       (generated with the {@link #Job constructor}), otherwise the numerical
       job ID (if generated by {@link #newInstance newInstance(id)}).
     */
    public int edgJobId() {
	return edgJobId;
    }

    private void setEdgJobId(int id) {
	edgJobId = id;
    }

    /**
       returns the job type.
       @return the numerical ID for the job type.
     */
    public int jobType() {
	return jobType;
    }	    

    /**
       notifies the job about to which
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       it has been submitted.
       @param ceID the numerical ID of the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}.
     */
    public void setExecutingCE(int ceID) {
	execCEId = ceID;
    }

    /**
       returns the ID of the executing
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}.
       @return {@link #NO_EXEC_CE Job.NO_EXEC_CE} if the job has not yet been
       submitted, otherwise the numerical ID of the executing
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
     */
    public int executingCE() {
	return execCEId;
    }

    /**
       returns the computational energy "consumed" by the job. The value is
       randomly chosen between <code>minCompEn</code> and
       <code>maxCompEn</code> upon job creation (see the
       {@link #Job constructor} or {@link #newInstance newInstance(id)}).
       @return the amount in <i>units of computational energy (UCE)</i>.
     */
    public int compEnergy() {
	return compEnergy;
    }

    /**
       returns the minimum processing power required by the job. The value is
       randomly chosen between <code>minCPUReq</code> and
       <code>maxCPUReq</code> upon job creation (see the
       {@link #Job constructor} or {@link #newInstance newInstance(id)}).
       @return the minimum processing power in <i>units of computational energy
       (UCE)</i> per <i>time unit</i>.
     */
    public int minCPUPower() {
	return minCPUPower;
    }

    /**
       verifies if a job of that type can be submitted to a specific queue type.
       @return <code>true</code> if the job can be submitted.
       @param qType the ID string of the queue type. Queue ID strings are
       defined in the configuration file.
     */
    public boolean qTypeAccepted(String qType) {
	if (requiredQTypes == null) // default, accept all
	    return true;

	return requiredQTypes.contains(qType);
    }

    // may be overwritten by derived classes!!!
    /**
       tries to match the characteristics of a specific
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       to the job's requirements. In this implementation the following
       conditions will be controlled: the minimum processing power, the queue
       type and the possibility to submit the job without exceeding the
       resource's maximum queue wait time (QWT). Derived classes might
       overwrite this method in order to add further conditions.
       @return <code>true</code> if the job matches the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       and can be submitted.
       @param ce a reference to the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}.
     */
    public boolean matchCE(ComputingElement ce) {

	if ( ce.getCPUPower() >= minCPUPower() // min. "quality" req.
	     && qTypeAccepted(ce.getQTypeIdString()) // queue type ok
	     && ce.jobMayBeQueued(this) // job still can be queued to CE.
	     ) {

	    return true;
	}

	return false;
    }

    /**
       returns the price that is charged for the execution of ONE <i>unit of
       computational energy (UCE)</i> of the job. To compute the full cost for
       job execution use
       <code><b>getJobPrice()*{@link #compEnergy compEnergy()}</b></code>.
       The price will be set using {@link #setJobPrice setJobPrice(price)}. 
       @return the price in GridCredits per <i>unit of computational energy
       (UCE)</i>.
     */
    public int getJobPrice() {
	return price;
    }

    /**
       notifies the job of the price that is charged for the execution of ONE
       <i>unit of computational energy (UCE)</i> by the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       to which it has been submitted.
       @param price the price in GridCredits per <i>unit of computational
       energy (UCE)</i>.
     */
    public void setJobPrice(int price) {
	this.price = price;
    }

    /**
       creates a new random job of this job type (see the 
       {@link #Job constructor}) setting the "consumed" computational energy to
       a random value between <code>minCompEn</code> and <code>maxCompEn</code>
       and the required minimum processing power to a random value between
       <code>minCPUReq</code> and <code>maxCPUReq</code>. The job type and the
       accepted queue types remain the same. The newly generated job's ID is
       set as specified.
       @return a reference to the newly generated job.
       @param id the numerical ID of the new job, should be unique.
     */
    public Job newInstance(int id) {
	Job newJob = new Job(jobType, minJobCompEn, maxJobCompEn,
			    minJobCPUReq, maxJobCPUReq, requiredQTypes);
	newJob.setEdgJobId(id);

	return newJob;
    }
}
