package org.edg.wl.dgas.dgassim.fabric;

import java.util.*;
import java.io.*;
import org.edg.wl.dgas.dgassim.*;
import org.edg.wl.dgas.dgassim.workload.pa.*;

/**
   simulates a Grid computing element (which might be a single worker node,
   a supercomputer or a cluster).
 */
public class ComputingElement {

    private boolean verbose;
    private boolean debug;

    private PriceAuthority pa;

    private int ceID;

    private String qTypeIdString;
    private int maxQWT; // maximum queue wait time in time units
    private int currQWT; // current queue wait time in time units

    private int cpuPower; // CPU power in UCE/time unit.

    /**
       defines the duration of a <i>time unit</i> in seconds.
     */
    public static double SEC_PER_TIME_UNIT;

    private long timeOfLastComp; // timestamp of last computation

    private int profit;

    // for data file creation:
    private BufferedWriter qwtOutFile, priceOutFile, profitOutFile;

    /**
       creates an instance.
       @param id a numerical ID for the computing element.
       @param secPerTU the number of seconds per <i>time unit</i> (see
       {@link #SEC_PER_TIME_UNIT SEC_PER_TIME_UNIT}). This parameter sets the
       global static value for all <code><b>ComputingElement</b></code>
       instances.
       @param pa the
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}
       to which to register the <code><b>ComputingElement</b></code> instance to.
       @param qTIdStr the ID string that describes the queue type for the
       <code><b>ComputingElement</b></code>. ID strings are defined in the
       configuration file.
       @param cpuP a numerical value (benchmark on a linear scale) for the
       processing power. The processing power is measured in <i>units of
       computational energy (UCE)</i> furnished per <i>time unit</i>.
       @param mQWT the maximum queue wait time (QWT) in <i>time units</i>.
       @param cQWT the current (initial) queue wait time (QWT) in <i>time
       units</i>.
       @param initialPrice the initial resource price. If negative the price
       will be computed by the
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}.
       Note: Some pricing algorithms may require an initial price to be set
       instead of being computed!
       @param verbose sets the verbose flag to print information to the
       standard output.
       @param debug sets the debug flag to print additional debugging
       information to the standard output.
     */
    public ComputingElement(int id, double secPerTU, PriceAuthority pa,
			    String qTIdStr,
			    int cpuP, int mQWT, int cQWT,
			    int initialPrice,
			    boolean verbose, boolean debug) {

	SEC_PER_TIME_UNIT = secPerTU;

	this.verbose = verbose;
	this.debug = debug;

	if (debug)
	    System.out.println("constructor of class ComputingElement");

	if (debug)
	    System.out.println("SEC_PER_TIME_UNIT = "+SEC_PER_TIME_UNIT);

	qTypeIdString = new String(qTIdStr);

	this.pa = pa;
	ceID = id;
	timeOfLastComp = (new Date()).getTime(); // current time
	maxQWT = mQWT;
	currQWT = cQWT;
	cpuPower = cpuP;

	profit = 0;

	// register CE to PA:
	if (pa != null) {
	    if (initialPrice < 0) // compute initial price
		pa.registerCE(this);
	    else // set initial price
		pa.registerCE(this, initialPrice);
	}
    } // constructor

    /**
       gets a reference to the output file to which the
       <code><b>ComputingElement</b></code>'s queue wait time (QWT) information
       has to be written.
       @return A reference to the <code><b>BufferedWriter</b></code> for QWT
       information.
     */
    public BufferedWriter getQWTOutFile() {
	return qwtOutFile;
    }
    /**
       gets a reference to the output file to which the
       <code><b>ComputingElement</b></code>'s price information has to be
       written.
       @return A reference to the <code><b>BufferedWriter</b></code> for
       price information.
     */
    public BufferedWriter getPriceOutFile() {
	return priceOutFile;
    }
     /**
       gets a reference to the output file to which the
       <code><b>ComputingElement</b></code>'s profit information has to
       be written. Note: Not all pricing methods use profit information,
       the output file may not exist or be required.
       @return A reference to the <code><b>BufferedWriter</b></code> for
       profit information.
     */
    public BufferedWriter getProfitOutFile() {
	return profitOutFile;
    }

    /**
       sets the output file to which the <code><b>ComputingElement</b></code>'s
       queue wait time (QWT) information has to be written.
       @param qwtOF a reference to the <code><b>BufferedWriter</b></code> for
       QWT information.
     */
    public void setQWTOutFile(BufferedWriter qwtOF) {
	qwtOutFile = qwtOF;
    }
    /**
       sets the output file to which the <code><b>ComputingElement</b></code>'s
       price information has to be written.
       @param pOF a reference to the <code><b>BufferedWriter</b></code> for
       price information.
     */
    public void setPriceOutFile(BufferedWriter pOF) {
	priceOutFile = pOF;
    }
    /**
       sets the output file to which the <code><b>ComputingElement</b></code>'s
       profit information has to be written. Note: Not all pricing methods use
       profit information, the output file may not be required.
       @param pOF a reference to the <code><b>BufferedWriter</b></code> for
       profit information.
     */
    public void setProfitOutFile(BufferedWriter pOF) {
	profitOutFile = pOF;
    }

    /**
       computes the current queue wait time (QWT) by simply subtracting the
       number of <i>time units</i> that have passed since the last computation.
       @see #queueJob queueJob()
     */
    public void computeJobsInQueue() {
	long currTimestamp = (new Date()).getTime();

	synchronized (this) {
	    int seconds = ( (int)(currTimestamp - timeOfLastComp) / 1000 );

	    if ((double)seconds >= SEC_PER_TIME_UNIT) {
		int timeUnits = (int)(seconds / SEC_PER_TIME_UNIT);
		// the CPU power was already considered when queueing the job!

		currQWT -= timeUnits; // "compute" the jobs in queue.

		if (currQWT < 0)
		    currQWT = 0;

		timeOfLastComp += (int) (timeUnits * SEC_PER_TIME_UNIT * 1000);
	    }
	}
    }

    /**
       verifies whether a given job may be queued or not. Calls
       {@link #computeJobsInQueue computeJobsInQueue()} before verification.
       @return <code>true</code> if the current QWT plus the jobs run time
       don't exceed the maximum QWT.
       @param job the {@link org.edg.wl.dgas.dgassim.Job Job} that should be
       queued.
       @see #queueJob queueJob()
    */
    public boolean jobMayBeQueued(Job job){

	computeJobsInQueue();

	int timeUnits = (int) ((double)job.compEnergy()/(double)cpuPower +0.5);
	// how long does the job take to compute?

	if ( (maxQWT - currQWT) >= timeUnits ) { // queue not yet full
	    return true;
	}

	return false; // job too long for queue.
    }

    /**
       queues a given job if the current QWT plus the jobs run time don't
       exceed the maximum QWT. Calls
       {@link #computeJobsInQueue computeJobsInQueue()} before trying to queue.
       The current QWT is increased by the job run time in <i>time units</i>,
       which is computed by dividing the job's required computational energy
       (see {@link org.edg.wl.dgas.dgassim.Job#compEnergy Job.compEnergy()})
       by the <code><b>ComputingElement</b></code>'s processing power (in
       <i>units of computational energy (UCE)</i> per <i>time unit</i>).
       @return <code>true</code> if the job could be queued.
       @param job the {@link org.edg.wl.dgas.dgassim.Job Job} that has to be
       queued.
    */
    synchronized public boolean queueJob(Job job){

	computeJobsInQueue();

	int timeUnits = (int) ( (double)job.compEnergy() / (double)cpuPower
				+ 0.5);              // how long does the job
	                                             // take to compute?
	if ( (maxQWT - currQWT) >= timeUnits ) { // queue not yet full
	    currQWT += timeUnits;

	    // increase profit (take the price from the job info instead
	    // of the PA, since it might have changed (threads!). Jobs pay
            // the price that was valid, when the CE was selected.
            // NOTE: using the profit of the WHOLE computational energy
            // causes far to big fluctuations, so we use the profit per
            // UCE (unit of comp. energy) which is equal to the price.
            profit += job.getJobPrice(); // deleted: * job.compEnergy();

	    job.setExecutingCE(ceID);

	    if (verbose)
		System.out.println("Queueing job " + job.edgJobId()+" of type "
				   + job.jobType() + ": " 
				   + job.compEnergy() +" UCE, req. CPU power: "
				   + job.minCPUPower() + "\n -> to CE" + ceID
				   + " with CPU power: " + cpuPower
				   + " UCE/time units.\n"
				   + " -> " + timeUnits + " time units added"
                                   + " to CE's queue. Profit is now "+profit);

	    return true;
	}

	return false; // not registered, queue too full.
    }

    /**
       gets the
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}
       to which the <code><b>ComputingElement</b></code> is registered.
       @return a reference to the
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}
       responsible for setting the <code><b>ComputingElement</b></code>'s price.
    */
    public PriceAuthority getPA() {
	return pa;
    }

    /**
       gets the current price of the resource by calling
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority#getPrice PriceAuthority.getPrice(ceID)}
       of the responsible
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}.
       @return the price in <i>GridCredits</i>.
    */
    public int getPrice() {
	return pa.getPrice(ceID);
    }

    /**
       gets the accumulated profit of the resource.
       @return the profit in <i>GridCredits</i>.
    */
    public int getProfit() {
	return profit;
    }

    /**
       sets the accumulated profit of the resource. Generally used to set the
       profit back to zero.
       @param newProfit the profit in <i>GridCredits</i>.
     */
    public void setProfit(int newProfit) {
	profit = newProfit;
    }

    /**
       gets a description of the <code><b>ComputingElement</b></code>'s queue
       type. The strings that describe the queue types are usually defined in a
       configuration file.
       @return a string describing the queue type.
    */
    public String getQTypeIdString() {
	return new String(qTypeIdString);
    }

    /**
       gets the current QWT of the <code><b>ComputingElement</b></code>. Calls
       {@link #computeJobsInQueue computeJobsInQueue()} before returning
       the current QWT.
       @return the current QWT in <i>time units</i>.
       @see #getCurrQWTNoComp getCurrQWTNoComp
    */
    synchronized public int getCurrQWT() {
	computeJobsInQueue();
	return currQWT;
    }

    /**
       gets the current QWT of the <code><b>ComputingElement</b></code> without
       calling {@link #computeJobsInQueue computeJobsInQueue()}.
       @return the (not necessarily accurate) current QWT in <i>time units</i>.
       @see #getCurrQWT getCurrQWT
    */
    synchronized public int getCurrQWTNoComp() {
	return currQWT;
    }

    /**
       gets the maximum QWT of the <code><b>ComputingElement</b></code>.
       @return the maximum QWT in <i>time units</i>.
     */
    synchronized public int getMaxQWT() {
	return maxQWT;
    }

    /**
       gets the processing power of the <code><b>ComputingElement</b></code>
       (in <i>units of computational energy (UCE)</i>
       per <i>time unit</i>).
       @return the processing power in <i>UCE/t.u.</i>
     */
    synchronized public int getCPUPower() {
	return cpuPower;
    }

    /**
       gets the <code><b>ComputingElement</b></code>'s ID.
       @return the numerical ID.
     */
    synchronized public int getId() {
	return ceID;
    }

}
