package org.edg.wl.dgas.dgassim.monitor;

import java.util.*;
import org.edg.wl.dgas.dgassim.*;
import org.edg.wl.dgas.dgassim.workload.pa.*;
import org.edg.wl.dgas.dgassim.fabric.*;

/**
   simulates the Grid Information System (<code><b>GIS</b></code>) by keeping
   references to all
   {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
   and {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}s.
   It additionally keeps references to all {@link org.edg.wl.dgas.dgassim.Job Job}
   types and queue types defined in the configuration file.
 */
public class GIS {

    private static int currEdgJobId = 0;

    private boolean verbose, debug;

    private ArrayList ceIDs;
    private TreeMap ces; // maps CE IDs to CE references.

    private ArrayList paIDs;
    private TreeMap pas; // maps PA IDs to PA references.
    private TreeMap paStr; // maps PA ID strings to PA references.

    private ArrayList jobTypeIDs;
    private TreeMap jobTypes; // maps job type IDs to job references.

    private TreeMap maxQWTs; // maps queue ID strings max. QWTs.

    /**
       creates an instance.
       @param verbose sets the verbose flag to print information to the
       standard output.
       @param debug sets the debug flag to print additional debugging
       information to the standard output.
     */
    public GIS(boolean verbose, boolean debug) {

	ces = new TreeMap();
	ceIDs = new ArrayList();

	pas = new TreeMap();
	paStr = new TreeMap();
	paIDs = new ArrayList();

	jobTypes = new TreeMap();
	jobTypeIDs = new ArrayList();

	maxQWTs = new TreeMap();

	this.verbose = verbose;
	this.debug = debug;

	if (debug)
	    System.out.println("constructor of class GIS");
    }

    /**
       gets the current edgJobId (unique numerical job ID). Job ID's are
       incremented by <code>1</code> each time a new simulated
       {@link org.edg.wl.dgas.dgassim.Job Job} is generated.
       @return the numerical ID of the last generated Job.
     */
    public int getCurrEdgJobId() {
	return currEdgJobId;
    }

    /**
       gets an <code><b>ArrayList</b></code> of references to all registered
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s.
       @return an <code><b>ArrayList</b></code> of references to
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s.
     */
    public ArrayList getCEList() {
        ArrayList list = new ArrayList();
	synchronized (this) {
	    for (int i = 0; i < ceIDs.size(); i++) {
		list.add((ComputingElement)ces.get(ceIDs.get(i)));
	    }
	}
        return list;
    }

    /**
       gets an <code><b>ArrayList</b></code> of references to all registered
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}s.
       @return an <code><b>ArrayList</b></code> of references to
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}s.
     */
    public ArrayList getPAList() {
        ArrayList list = new ArrayList();
	synchronized (this) {
	    for (int i = 0; i < paIDs.size(); i++) {
		list.add((PriceAuthority)pas.get(paIDs.get(i)));
	    }
	}
        return list;
    }

    /**
       gets an <code><b>ArrayList</b></code> of references to all registered
       {@link org.edg.wl.dgas.dgassim.Job Job} types.
       @return an <code><b>ArrayList</b></code> of references to
       {@link org.edg.wl.dgas.dgassim.Job Job}s.
     */
    public ArrayList getJobTypeList() {
        ArrayList list = new ArrayList();
	synchronized (this) {
	    for (int i = 0; i < jobTypeIDs.size(); i++) {
		list.add((Job)jobTypes.get(jobTypeIDs.get(i)));
	    }
	}
        return list;
    }

    /**
       gets an <code><b>ArrayList</b></code> of <code><b>Integer</b></code>s
       containing the numerical IDs of all registered
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s.
       @return  an <code><b>ArrayList</b></code> of <code><b>Integer</b></code>s.
     */
    public ArrayList getCEIDList() {
	return ceIDs;
    }

    /**
       gets an <code><b>ArrayList</b></code> of <code><b>Integer</b></code>s
       containing the numerical IDs of all registered
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}s.
       @return  an <code><b>ArrayList</b></code> of <code><b>Integer</b></code>s.
     */
    public ArrayList getPAIDList() {
	return paIDs;
    }

    /**
       gets an <code><b>ArrayList</b></code> of <code><b>Integer</b></code>s
       containing the numerical IDs of all registered
       {@link org.edg.wl.dgas.dgassim.Job Job} types.
       @return  an <code><b>ArrayList</b></code> of <code><b>Integer</b></code>s.
     */
    public ArrayList getJobTypeIDList() {
	return jobTypeIDs;
    }

    /**
       gets the registered
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       having a certain ID.
       @return a reference to the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}.
       @param ceID the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}'s
       numerical ID.
     */
    public ComputingElement getCE(int ceID) {
	return (ComputingElement) ces.get(new Integer(ceID));
    }

    /**
       gets the registered
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}
       having a certain ID.
       @return a reference to the
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}.
       @param paID the
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}'s
       numerical ID.
     */
    public PriceAuthority getPA(int paID) {
	return (PriceAuthority) pas.get(new Integer(paID));
    }

    /**
       gets the registered
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}
       having a certain ID string (defined in the configuration file).
       @return a reference to the
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}.
       @param paIDStr the
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}'s
       ID string.
     */
    public PriceAuthority getPA(String paIDStr) {
	return (PriceAuthority) paStr.get(paIDStr);
    }

    /**
       gets the registered
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}
       that is responsible for setting the price of a certain
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}.
       @return a reference to the
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}.
       @param ceId the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}'s
       numerical ID.
     */
   public PriceAuthority getCEsPA(int ceId) {
	return ((ComputingElement)ces.get(new Integer(ceId))).getPA();
    }

    /**
       gets the price of a registered
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       by querying the responsible
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}.
       @return the price in <i>GridCredits</i> per <i>unit of computational
       energy (UCE)</i>.
       @param ceId the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}'s
       numerical ID.
     */
    public int getPrice(int ceId) {
	return getCEsPA(ceId).getPrice(ceId);
    }

    /**
       gets the registered {@link org.edg.wl.dgas.dgassim.Job Job} type having
       a certain ID.
       @return a reference to the {@link org.edg.wl.dgas.dgassim.Job Job} type.
       @param jobTypeID the {@link org.edg.wl.dgas.dgassim.Job Job} type's
       numerical ID.
     */
    public Job getJobType(int jobTypeID) {
	return (Job) jobTypes.get(new Integer(jobTypeID));
    }

    /**
       creates a new job instance of the registered
       {@link org.edg.wl.dgas.dgassim.Job Job} type having a certain ID.
       @return a reference to the newly created
       {@link org.edg.wl.dgas.dgassim.Job Job}.
       @param jobTypeID the {@link org.edg.wl.dgas.dgassim.Job Job} type's
       numerical ID.
     */
    public Job createNewJob(int jobTypeID) {
	return getJobType(jobTypeID).newInstance(++currEdgJobId);
    }

    /**
       creates a new job instance randomly choosing one of the registered
       {@link org.edg.wl.dgas.dgassim.Job Job} types.
       @return a reference to the newly created
       {@link org.edg.wl.dgas.dgassim.Job Job}.
     */
    public Job createNewJob() {
	// create one at random:
	int jobNum = (int) ( (double)jobTypeIDs.size() *
			                          (double)Math.random() );
	return createNewJob(jobNum);
    }

    /**
       verifies whether a specified queue type is registered to this
       <code><b>GIS</b></code>.
       @return <code>true</code> if registered, otherwise <code>false</code>.
       @param qTypeStr an ID string for the queue type (ID strings are defined
       in the configuration file).
     */
    public boolean queueIsRegistered(String qTypeStr) {
	return maxQWTs.containsKey(qTypeStr);
    }

    /**
       gets the maximum queue wait time (QWT) of a registered queue type.
       @return the maximum QWT in <i>time units</i>.
       @param qTypeStr an ID string for the queue type (ID strings are defined
       in the configuration file).
     */
    public int getMaxQWT(String qTypeStr) {
	return ((Integer)maxQWTs.get(qTypeStr)).intValue();
    }

    /**
       registers a
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       to this <code><b>GIS</b></code>. If a previously registered
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       has the same numerical ID, then it will be replaced by the new one.
       @param ce the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       to be registered.
     */
    public void registerCE(ComputingElement ce) {
	synchronized (this) {
	    Integer id = new Integer(ce.getId());
	    ces.put(id, ce);
	    if (!ceIDs.contains(id))
		ceIDs.add(id);
	}
    }

    /**
       registers a
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}
       to this <code><b>GIS</b></code>. If a previously registered
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}
       has the same numerical ID, then it will be replaced by the new one.
       @param pa the
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}
       to be registered.
     */
    public void registerPA(PriceAuthority pa) {
	synchronized (this) {
	    Integer id = new Integer(pa.getId());
	    pas.put(id, pa);
	    paStr.put(pa.getIdString(), pa);
	    if (!paIDs.contains(id))
		paIDs.add(id);
	}
    }

    /**
       registers a {@link org.edg.wl.dgas.dgassim.Job Job} type to this
       <code><b>GIS</b></code>. If a previously registered
       {@link org.edg.wl.dgas.dgassim.Job Job} type has the same numerical ID,
       then it will be replaced by the new one.
       @param jType the {@link org.edg.wl.dgas.dgassim.Job Job} type to be
       registered.
    */
    public void registerJobType(Job jType) {
	synchronized (this) {
	    Integer id = new Integer(jType.jobType());
	    jobTypes.put(id, jType);
	    if (!jobTypeIDs.contains(id))
		jobTypeIDs.add(id);
	}
    }

    /**
       registeres a queue type to this <code><b>GIS</b></code>. If a previously
       registered queue type has the same ID string (defined in the
       configuration file), then it will be replaced by the new one.
       @param qTypeStr the ID string of the queue type.
       @param maxQWT the maximum queue wait time (QWT) in <i>time units</i> to
       be mapped to the ID string.
     */
    public void registerQType(String qTypeStr, int maxQWT) {
	synchronized (this) {
	    maxQWTs.put(new String(qTypeStr), new Integer(maxQWT));
	}
    }

}
