package org.edg.wl.dgas.dgassim.workload.pa;

import org.edg.wl.dgas.dgassim.*;
import org.edg.wl.dgas.dgassim.fabric.*;

/**
   implements the Hybrid Pricing Model as a pricing module for
   {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}s.
   This implementation is identical to the description of the model presented
   in: R. Piro, A. Guarise, and A. Werbrouck, "An Economy-based Accounting
   Infrastructure for the DataGrid", <i>Proceedings of the 4th International
   Workshop on Grid Computing (GRID2003)</i>, held in conjunction with
   SuperComputing 2003, Phoenix, Arizona, November 17, 2003.
 */
public class HybridPricingModule implements PricingModule {

    private boolean verbose;
    private boolean debug;

    private int basePrice, deltaP;

    /**
       creates an instance to be used for a
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}.
       @param basePrice the base price in <i>GridCredits</i> per <i>unit of
       computational energy (UCE)</i> that will be charged for all resources
       that are registered to the
       {@link org.edg.wl.dgas.dgassim.workload.pa.PriceAuthority PriceAuthority}.
       @param deltaP the price variation limit. Prices will vary between
       <code>basePrice-deltaP</code> and <code>basePrice+deltaP</code>.
       @param verbose sets the verbose flag to print information to the
       standard output.
       @param debug sets the debug flag to print additional debugging
       information to the standard output.
    */
    public HybridPricingModule(int basePrice, int deltaP,
			       boolean verbose, boolean debug) {

	this.basePrice = basePrice;
	this.deltaP = deltaP;

	this.verbose = verbose;
	this.debug = debug;

	if (debug)
	    System.out.println("constructor of class HybridPricingModule");

	if (debug)
	    System.out.println("BASE_PRICE="+basePrice
			       +", DELTA_P="+deltaP);

    }

    /**
       computes a new price based on the current queue wait time (cQWT) with
       respect to the maximum QWT (mQWT) for a particular resource.<br>
       Formula:<br>
       <code>price = basePrice + deltaP*( (cQWT - mQWT/2) / (mQWT/2) ) )</code>
       @return the price in <i>GridCredits</i> per <i>unit of computational
       energy (UCE)</i>.
       @param ce the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       for which to compute the new price.
     */
    public int computePrice(ComputingElement ce) {

	double cQWT = (double) ce.getCurrQWT();
	double mQWT = (double) ce.getMaxQWT();

	int price = (int) (basePrice +
			   deltaP*( (cQWT - mQWT/2) / (mQWT/2) ) );

	if (debug)
	    System.out.println("[HybridPricingModule] ceID="+ce.getId()
			       +", cQWT="+cQWT+", mQWT="
			       +mQWT+" => new price is "+price);

	return price;
    }

}
