package org.edg.wl.dgas.dgassim.workload.pa;

import java.util.*;
import java.io.*;
import org.edg.wl.dgas.dgassim.*;
import org.edg.wl.dgas.dgassim.fabric.*;

/**
   simulates a Price Authority (PA) of the
   <i><a href="http://www.to.infn.it/grid/accounting/">DataGrid Accounting
   System (DGAS)</a></i> that is part of the workload management system of the
   <i><a href="http://www.edg.org/">European DataGrid (EDG)</a></i> project.
   Each PA is responsible for setting the prices of a subset of Grid resources
   that is usually linked to a <i>Virtual Organization</i>. The
   <code><b>PriceAuthority</b></code> class extends <code><b>Thread</b></code>
   in order to allow for asynchronous pricing as it is done by the respective
   DGAS component. For more detailed documentation on PAs see: A. Guarise,
   A. Werbrouk and R. Piro. "DataGrid Accounting System - Architecture v1.0",
   Technical Report, DataGrid-01-TED-0126-1_0, INFN Turin, Italy, February 2003.
 */
public class PriceAuthority extends Thread {

    private boolean verbose, debug;

    private int paID;
    private String paIDString;

    private ArrayList ceIDs;
    private TreeMap prices; // maps CE IDs to prices.
    private TreeMap ces; // maps CE IDs to CE references.

    private boolean runPA;

    private PricingModule pmod;

    private int PA_WAIT_TIME;
    private int PRICE_ADJ_THRES;

    private boolean OUTPUT_IF_PRICE_NOT_ADJ;

    private long timeOfLastComp; // timestamp of last price computation
    private long timeOfFirstComp; // timestamp of first price computation

    /**
       will be returned instead of a price if the
       <code><b>PriceAuthority</b></code> is queried for the price of a
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       that is not registered. The value of
       <code><b>CE_NOT_REGISTERED</b></code> is <code><b>Integer.MIN_VALUE</b>
       (-2147483648)</code>.
     */
    public static final int CE_NOT_REGISTERED = Integer.MIN_VALUE;

    /**
       creates an instance that simulates a DGAS Price Authority server.
       @param id a numerical ID that schould be unique.
       @param paIdStr a describing ID string that should be unique. ID strings
       are defined in the configuration file.
       @param pm the {@link org.edg.wl.dgas.dgassim.workload.pa.PricingModule
       PricingModule} to be used for pricing the registered resources.
       @param paWaitTime the time interval (in miliseconds) between price
       adjustments. It can be considered a wait time or a time-to-live after
       which new prices will be computed for all registered resources.
       @param priceAdjThres the threshold for price adjustments. A new price
       is set only its difference to the previous price is significant.
       @param outputIfPriceNotAdj if <code>false</code> the output will be
       written only when prices are adjusted (see <code>priceAdjThres</code>).
       If <code>true</code> the output will be written, even if prices are not
       adjusted because the difference doesn't exceed the threshold.
       @param verbose sets the verbose flag to print information to the
       standard output.
       @param debug sets the debug flag to print additional debugging
       information to the standard output.
     */
    public PriceAuthority(int id, String paIdStr,
			  PricingModule pm, int paWaitTime,
			  int priceAdjThres, boolean outputIfPriceNotAdj,
			  boolean verbose, boolean debug) {
	super();

	paID = id;
	paIDString = new String(paIdStr);

	PA_WAIT_TIME = paWaitTime;
	PRICE_ADJ_THRES = priceAdjThres;
	OUTPUT_IF_PRICE_NOT_ADJ = outputIfPriceNotAdj;

	this.verbose = verbose;
	this.debug = debug;

	if (debug)
	    System.out.println("constructor of class PriceAuthority:\n"
			       + "PA_WAIT_TIME=" + PA_WAIT_TIME
			       + ", PRICE_ADJ_THRES=" + PRICE_ADJ_THRES);

	prices = new TreeMap();
	ces = new TreeMap();
	ceIDs = new ArrayList();
	pmod = pm;
	timeOfLastComp = (new Date()).getTime(); // current time
	timeOfFirstComp = timeOfLastComp;
    }

    /**
       gets the <code><b>PriceAuthority</b></code>'s ID.
       @return the numerical ID.
     */
    public int getId() {
	return paID;
    }

    /**
       gets the <code><b>PriceAuthority</b></code>'s decriptive ID.
       @return the ID string as defined in the configuration file.
     */
    public String getIdString() {
	return new String(paIDString);
    }

    /**
       queries the <code><b>PriceAuthority</b></code> for price information on
       a specific
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}.
       Since each <code><b>PriceAuthority</b></code> instance is responsible
       only for a subset of Grid resources, it is recommended to use the method
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement#getPrice ComputingElement.getPrice()}
       instead.
       @return the price in <i>GridCredits</i> per <i>unit of computational
       energy (UCE)</i> or
       {@link #CE_NOT_REGISTERED PriceAuthority.CE_NOT_REGISTERED} if the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       is not registered to this <code><b>PriceAuthority</b></code> instance.
       @param ceId the {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement
       ComputingElement}'s numerical ID.
     */
    synchronized public int getPrice(int ceId) {
	Integer price = (Integer)prices.get(new Integer(ceId));

	if (price == null)
	    return CE_NOT_REGISTERED;

	return price.intValue();
    }

    /**
       registers a
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       to this <code><b>PriceAuthority</b></code> instance and sets a specified
       initial price for it. If a previously registered
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       has the same numerical ID, then it will be replaced by the new one.
       @param ce the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       to be registered.
       @param price the initial price in <i>GridCredits</i> per <i>unit of
       computational energy (UCE)</i>.
     */
    synchronized public void registerCE (ComputingElement ce, int price) {
	Integer id = new Integer(ce.getId());

	prices.put(id, new Integer(price));
	ces.put(id, ce);
	if (!ceIDs.contains(id))
	    ceIDs.add(id);
    }
    // if the price is not specified
    /**
       registers a
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       to this <code><b>PriceAuthority</b></code> instance and lets the
       {@link PricingModule PricingModule} set an initial price for it. If a
       previously registered
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       has the same numerical ID, then it will be replaced by the new one.
       Note: Not all {@link PricingModule PricingModule}s can determine initial
       prices, in this case {@link #registerCE registerCE(ce, price)} has to be
       used to register the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}.
       @param ce the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}
       to be registered.
     */
    synchronized public void registerCE (ComputingElement ce) {
	registerCE(ce, pmod.computePrice(ce));
    }

    /**
       stops this <code><b>Thread</b></code>. Note: The class doesn't implement
       the method <code><b>Thread.stop()</b></code> but instead keeps a flag
       (<code>runPA</code>) that is periodically checked in {@link #run run()}.
       This method sets <code>runPA</code> to <code>false</code>.
     */
    public void stopPA() {
	runPA = false;
    }

    /**
       inherited from <code><b>Thread</b></code>. Starts this
       <code><b>PriceAuthority</b></code> instance.
    */
    public void start() {
	runPA = true;
	if (verbose) System.out.println("Starting PA thread: " + paIDString);
	super.start();
    }

    /**
       inherited from <code><b>Thread</b></code>. Periodically computes new
       prices for all registered
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
       using the {@link PricingModule PricingModule} specified during the
       creation of this <code><b>PriceAuthority</b></code> instance (see the
       {@link #PriceAuthority constructor}). Note: The class doesn't implement
       the method <code><b>Thread.stop()</b></code> but instead keeps a flag
       (<code>runPA</code>) that is periodically checked in
       <code><b>run()</b></code>. A while-loop is executed as long as
       <code>runPA</code> is <code>true</code>.
     */
    public void run() {
	long secSinceFirst;

	BufferedWriter currFile;
	String currLine;

	ComputingElement currCE;
	Integer currCEId;

	while(runPA) {
	    // compute new prices if waitTime is over ...
	    long currTimestamp = (new Date()).getTime();

	    if (currTimestamp - timeOfLastComp >= PA_WAIT_TIME) {

		secSinceFirst = (currTimestamp-timeOfFirstComp)/1000;

		if (verbose)
		    System.out.println("[PA" + paID + " (" + paIDString + ")] "
				       + "sec=" + secSinceFirst + ", last="
				       + timeOfLastComp + ", now="
				       + currTimestamp);

		boolean writeOutput;
		int size = ceIDs.size();
		int newPrice;
		int oldPrice;
		for (int i = 0; i < size; i++) {

		    currCEId = (Integer)ceIDs.get(i);
		    currCE = (ComputingElement)ces.get(currCEId);
		    oldPrice = getPrice(currCEId.intValue());
		    newPrice = pmod.computePrice(currCE);

		    if (verbose)
			System.out.println("[PA" + paID + " (" + paIDString
					   + ")] CE" + currCEId
					   + ": oldPrice=" + oldPrice
					   + ", newPrice=" + newPrice
					   + ", threshold=" + PRICE_ADJ_THRES);

		    writeOutput = true;
		    // verify if price adjustment threshold is exceeded
		    if ( (newPrice - oldPrice) >= PRICE_ADJ_THRES ||
			 (newPrice - oldPrice) <= -PRICE_ADJ_THRES ) {

			if (verbose) System.out.println(" -> ... price "
							+ "set.");
			prices.put(currCEId, new Integer(newPrice));
		    }
		    else { // doesn't exceed threshold
			if (verbose) System.out.println(" -> ... price NOT"
							+ " set.");
			newPrice = ((Integer)prices.get(currCEId)).intValue();

			// verify whether to write the output
			if (! OUTPUT_IF_PRICE_NOT_ADJ)
			    writeOutput = false;
		    }


		    synchronized(this) {
			if (writeOutput) {
			    // write to output files:
			    try {
				// all the price information
				currFile = currCE.getPriceOutFile();
				currLine = "" + secSinceFirst + " "
				    + newPrice+"\n";
				currFile.write(currLine);
				currFile.flush();
				// all the qwt information
				currFile = currCE.getQWTOutFile();
				currLine = "" + secSinceFirst + " "
				    + currCE.getCurrQWTNoComp() +"\n";
				currFile.write(currLine);
				currFile.flush();
			    } catch (Exception e) {
				System.err.println("\nCannot write data: "
						   + e.getMessage());
			    }
			}
		    } // synchronized ...

		} // for

		timeOfLastComp = (new Date()).getTime();
	    } // if ...

	    if (!runPA) {
		try {
		    // close output streams:
		    int size = ceIDs.size();
			for (int i = 0; i < size; i++) {
			    currCEId = (Integer)ceIDs.get(i);
			    currCE = (ComputingElement)ces.get(currCEId);
			    currFile = currCE.getPriceOutFile();
			    currFile.close();
			    currFile = currCE.getQWTOutFile();
			    currFile.close();
			}
		}
		catch (Exception ex) {
		    System.err.println("Cannot close price and QWT output"
				       + " files.");
		}
	    }

	} // while

	return;
    }
}
