package org.edg.wl.dgas.dgassim.workload.rb;

import java.util.*;
import org.edg.wl.dgas.dgassim.*;
import org.edg.wl.dgas.dgassim.fabric.*;

/**
   implements a utility module used by the
   {@link org.edg.wl.dgas.dgassim.workload.rb.ResourceBroker ResourceBroker}
   for scheduling decisions. The <code><b>PriceSensitiveUtilityModule</b></code>
   simply chooses the cheapest
   {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}. If
   a set of resources with equal low prices is available the module randomly
   chooses one for job submission.
 */
public class PriceSensitiveUtilityModule implements UtilityModule {

    /**
       creates an instance.
     */
    public PriceSensitiveUtilityModule() {
	// nothing to do ...
    }

    /**
       selects one of the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
       for job submission (see the
       {@link PriceSensitiveUtilityModule general class description}). The
       {@link org.edg.wl.dgas.dgassim.workload.rb.ResourceBroker ResourceBroker}
       will first use a
       {@link org.edg.wl.dgas.dgassim.workload.rb.Matchmaker Matchmaker} to get
       a set of matching resources for the given
       {@link org.edg.wl.dgas.dgassim.Job Job}.
       @return a reference to the selected
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}.
       @param ces an <code><b>ArrayList</b></code> of references to all matching
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
       as returned by the
       {@link org.edg.wl.dgas.dgassim.workload.rb.Matchmaker Matchmaker}.
       @param job the {@link org.edg.wl.dgas.dgassim.Job Job} for which to
       select a suitable resource.
     */
    public ComputingElement selectCE(ArrayList ces, Job job) {

	ArrayList selectedCEs = new ArrayList();
	ComputingElement ce = null;
	int price = Integer.MAX_VALUE;

	for (int i = 0; i < ces.size(); i++) {

	    ce = (ComputingElement)ces.get(i);

	    if ( price > ce.getPA().getPrice(ce.getId()) ) {
		// found a cheeper resource:
		selectedCEs.clear();
		price = ce.getPA().getPrice(ce.getId());
		selectedCEs.add(ce);
	    }
	    else if ( price == ce.getPA().getPrice(ce.getId()) ) {
		// found a CE with the same cheap price:
		selectedCEs.add(ce);
	    }
	}

	if (selectedCEs.size() < 1) {
	    return null; // no CE found
	}

	// if we found at least one CE
	int ranIndex = (int) (((double)Math.random())*selectedCEs.size());
	return (ComputingElement) selectedCEs.get(ranIndex);
    }

}
