package org.edg.wl.dgas.dgassim.workload.rb;

import java.util.*;
import org.edg.wl.dgas.dgassim.*;
import org.edg.wl.dgas.dgassim.fabric.*;

/**
   implements a utility module used by the
   {@link org.edg.wl.dgas.dgassim.workload.rb.ResourceBroker ResourceBroker}
   for scheduling decisions. The
   <code><b>RandPriceSensitiveUtilityModule</b></code> randomly chooses one
   of the <i>n</i> (={@link #numberOfCheapCEs numberOfCheapCEs}) cheapest
   {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s.
*/
public class RandPriceSensitiveUtilityModule implements UtilityModule {

    /**
       The number of cheap resources of which one will be randomly chosen.
       (see attribute NUMBER_OF_CHEAP_CES in the configuration file).
     */
    protected int numberOfCheapCEs;

    /**
       The default number of cheap resources, of which one will be
       randomly chosen, is <code>2</code>.
     */
    public static final int DEF_NUMBER_OF_CHEAP_CES = 2;

    /**
       creates an instance.
       @param nOCCEs the number of cheapest
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
       that have to be considered for random job scheduling (sets
       {@link #numberOfCheapCEs numberOfCheapCEs}).
     */
    public RandPriceSensitiveUtilityModule(int nOCCEs) {
	numberOfCheapCEs = nOCCEs;
    }

    /**
       creates an instance. {@link #numberOfCheapCEs numberOfCheapCEs} will be
       set to its default value of
       {@link #DEF_NUMBER_OF_CHEAP_CES DEF_NUMBER_OF_CHEAP_CES}.
     */
    public RandPriceSensitiveUtilityModule() {
	numberOfCheapCEs = DEF_NUMBER_OF_CHEAP_CES;
    }

    /**
       selects one of the
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
       for job submission (see the
       {@link PriceSensitiveUtilityModule general class description}). The
       {@link org.edg.wl.dgas.dgassim.workload.rb.ResourceBroker ResourceBroker}
       will first use a
       {@link org.edg.wl.dgas.dgassim.workload.rb.Matchmaker Matchmaker} to get
       a set of matching resources for the given
       {@link org.edg.wl.dgas.dgassim.Job Job}.
       @return a reference to the selected
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}.
       @param ces an <code><b>ArrayList</b></code> of references to all matching
       {@link org.edg.wl.dgas.dgassim.fabric.ComputingElement ComputingElement}s
       as returned by the
       {@link org.edg.wl.dgas.dgassim.workload.rb.Matchmaker Matchmaker}.
       @param job the {@link org.edg.wl.dgas.dgassim.Job Job} for which to
       select a suitable resource.
     */
    public ComputingElement selectCE(ArrayList ces, Job job) {

	// if there are no matching CEs
	if (ces.size() < 1) {
	    return null; // no CE found
	}

	ComputingElement ce = null;

	// first sort the CEs according to their price:

	TreeMap sortedCEsMap = new TreeMap();

	for (int i = 0; i < ces.size(); i++) {

	    ce = (ComputingElement)ces.get(i);

	    sortedCEsMap.put(new Integer(ce.getPrice()), ce);
	}

	// get the cheapest CEs:
	ArrayList selectedCEs = new ArrayList();
	while ( (sortedCEsMap.size() > 0) &&
		(selectedCEs.size() < numberOfCheapCEs) ) {
	    // add first element to selectedCEs
	    selectedCEs.add(sortedCEsMap.get(sortedCEsMap.firstKey()));
	    // remove first element from map
	    sortedCEsMap.remove(sortedCEsMap.firstKey());
	}

	// if we found at least one CE, randomly choose one
	int ranIndex = (int) (((double)Math.random())*selectedCEs.size());
	return (ComputingElement) selectedCEs.get(ranIndex);
    }

}
